#include "hbp_CommCoord.h"
#include <hpcxx_rts.h>		// for hpcxx_debug_level
#include <HPCxx_Thread.h>
#include <hpcxx_rtsImpl.h>	// for hpcxx_threadID()
#include <iostream.h>

//
// This class is designed to have a separate queue for each kernel level
// thread within a context.  Since the user-level threads are non-preemptive,
// only one user-level thread at a time can access the CommCoord 
// class methods.  The bottom line: no locks are needed to protect the
// communication object queues.
//

static const int NumDefaultCommObjects = 1;	// initial # of comm objects

//
// Class objects and methods (static) and inline functions
//

//---- hbp_CommCoord::receiveCommunicator -------------------------------------
// Returns the comm object corresponding to this receive page and offset.
//-----------------------------------------------------------------------------
inline
hbp_RecvComm* hbp_CommCoord::receiveCommunicator(unsigned int recv_pg,
						 unsigned int recv_off)
{
  void* buf = hbp_rcv_buf_ptr(hbp, recv_pg, recv_off);
  int sender = hbp_Comm::sender(buf);
  int index = hbp_Comm::index(buf);
    
  hbp_RecvComm* comm = recvQ[sender][index];
  comm->setInUse(true);

  return comm;
}


//---- hbp_CommCoord ----------------------------------------------------------
// Constructor.
//-----------------------------------------------------------------------------
hbp_CommCoord::hbp_CommCoord(int context, int nContexts,
			       char* hosts[], int nPages)
{
  int i;

  this->context = context;
  this->nContexts = nContexts;

  //
  // initialize hbp layer
  //

  initializeHBP(hosts);

  //
  // create communication queues
  //
  
  nComms = new int[nContexts];
  nextSend = new int[nContexts];

  for (i = 0; i < nContexts; i++) {
    nComms[i] = NumDefaultCommObjects;
    nextSend[i] = 0;
  }

  sendQ = new vector<hbp_SendComm*>[nContexts];
  recvQ = new vector<hbp_RecvComm*>[nContexts];

  // create communication objects and place in queues
  
  unsigned int dst_pg  = context / CONTEXTS_PER_PAGE;
  unsigned int dst_off = context % CONTEXTS_PER_PAGE;;

  for (i = 0; i < nContexts; i++) {
    for (int j = 0; j < NumDefaultCommObjects; j++) {
      int n = j + i*NumDefaultCommObjects;
      unsigned int pg = n / CONTEXTS_PER_PAGE;
      unsigned int pg_off = 2 * (n % CONTEXTS_PER_PAGE);

      // cmd/ack pair (for sending)

      void* buf = hbp_snd_buf_ptr(hbp, pg, pg_off);	// send commands
      CommBuffer* cmdBuf = new PageBuffer(CMD_BUF_SIZE, buf, pg, pg_off);

      buf = hbp_rcv_buf_ptr(hbp, pg, pg_off+1);	       // recv acks
      CommBuffer* ackBuf = new PageBuffer(CMD_BUF_SIZE, buf, pg, pg_off+1);

      HPCxx_Comm* comm = new bhp_SendComm(mutex_, bhp_, dst_pg, dst_off,
					  dst_port, cmdBuf, ackBuf);
      comm->setSender(i);
      comm->setSenderIndex(j);
      comm->setInUse(false);

      sendQ[i].push_back(comm);

      // cmd/ack pair (for receiving)

      buf = hbp_rcv_buf_ptr(hbp, pg, pg_off);		// recv commands
      cmdBuf = new PageBuffer(CMD_BUF_SIZE, buf, pg, pg_off);

      buf = hbp_snd_buf_ptr(hbp, pg, pg_off+1);		// send acks
      ackBuf = new PageBuffer(CMD_BUF_SIZE, buf, pg, pg_off+1);

      comm = new bhp_RecvComm(mutex_, bhp_, dst_pg, dst_off, dst_port,
			      cmdBuf, ackBuf);
      comm->setSender(i);
      comm->setSenderIndex(j);
      comm->setInUse(false);

      recvQ[i].push_back(comm);
    }
  }
}


//---- ~hbp_CommCoord ---------------------------------------------------------
// Destructor.  Frees all released communication objects.
//-----------------------------------------------------------------------------
hbp_CommCoord::~hbp_CommCoord(void)
{
  delete [] nComms;
  delete [] nextSend;
  delete [] sendQ;
  delete [] recvQ;
}


void hbp_CommCoord::initializeHBP(int nHosts, char* dst_hosts[])
{
  char          src_host[128]; // source HiPPI 800 host (board) name
  char          dst_host[128]; /* destination HiPPI 800 host (board) name    */
  char         *dst_hosts[1];  /* array of destination HiPPI 800 host names  */
  unsigned int  buf_pgs;       /* number of pages in send or receive buffer  */
  int           auth[3];       /* authentication key                         */

  int  i;
  char machine[128];


  if (argc != 3) {
    fprintf(stderr, "Usage:  %s <dest machine> <dest port>\n", argv[0]);
    return(NULL);
  }

  if (gethostname(machine, 128) < 0) {           /* get name of this machine */
    fprintf(stderr, "gethostname() failed.\n");
    return(NULL);
  }

  buf_pgs = (TEST_MSG_LEN + hbp_get_pagesize() - 1) / hbp_get_pagesize();
  if (buf_pgs == 0) buf_pgs = 1;

  auth[0] = getuid();                            /* likely to be unique      */
  auth[1] = getgid();
  auth[2] = getuid() + ~getgid();

  sprintf(src_host, "%s%s", machine, TEST_PIHOST_SUFFIX);
  sprintf(dst_host, "%s%s", argv[1], TEST_POHOST_SUFFIX);
  dst_hosts[0] = dst_host;

  fprintf(stdout, "Checking network connectivity ... ");  fflush(stdout);
  if (hbp_ping(src_host, dst_host, 1) < 0) {     /* check connectivity       */
    fprintf(stdout, "Failed.\n");
    fprintf(stderr, "WARNING:  No path seems to exist from %s \n", src_host);
    fprintf(stderr, "                                   to %s.\n", dst_host);
    fprintf(stderr, "          But going ahead with the test anyways ....\n");
  }
  else
    fprintf(stdout, "Passed.\n");

  for (i = 0; i < TEST_DEVS; i++) {
    hbps = hbp_open(pidevs[i], src_host, dst_hosts, 1,
                    buf_pgs, 1, buf_pgs, 1, auth);
    if (hbps != NULL) break;
  }

  if (hbps == NULL) {
    fprintf(stderr, "Failed to open bypass endpoint.\n");
    return(NULL);
  }

  hbp_print_state(hbps);

  return(hbps);  
}



//---- communicator -----------------------------------------------------------
// Returns a communications object.
//-----------------------------------------------------------------------------
HPCxx_Comm* hbp_CommCoord::communicator(const HPCxx_ContextID& dest,
					int msgType, bool unique)
{
  // Note that threads are assumed to be nonpreemptive so that it can be
  // assumed that only one user level thread will be running this at a time.
  // Although one or more threads may be blocking at the yield pt.

  hbp_SendComm* comm = sendQ[dest][nextSend[dest]];

  while ( comm->inUse() ) {
    HPCxx_Thread::yield();	// wait until this comm has been ack'd
  }
  comm->setInUse(true);

  nextSend[dest] = (1 + nextSend[dest]) % nComms[dest];

  return comm;
}


//---- poll -------------------------------------------------------------------
// Polls communication queues for events.
//-----------------------------------------------------------------------------
void hbp_CommCoord::poll(void)
{
  ssize_t size;
  unsigned in recv_pg, recv_off;

  int id = *(hpcxx_currentContext->getContextID());
  if (hpcxx_debug_level > 3) {
    cerr << id <<":Entering hbp_CommCoord::poll()" << endl;
  }

  //
  // Check for incoming messages.
  //
  while ( (size = hbp_readz(hbp, &recv_pg, &recv_off)) >= 0 ) {
    if (hpcxx_debug_level > 2) {
      cerr << id << ":HPCxx_CommCoord::poll incoming:\n  " << *comm << endl;
    }

    handleMessage( receiveCommunicator(recv_pg, recv_off) );
  }

  if (hpcxx_debug_level > 3) {
    cerr << id << ":Exiting HPCxx_CommCoord::poll()" << endl;
  }
}


//---- handleMessage ----------------------------------------------------------
// Handles an incoming message.
//-----------------------------------------------------------------------------
void hbp_CommCoord::handleMessage(HPCxx_Comm& comm)
{
  int msgType = comm.messageType();

  switch (msgType) {
    case AckCmdMsg :
      comm.release();
      break;
    case AckReplyMsg :
      hpcxx_returnHandler(comm);
      break;
    case AskForDataMsg:
      hpcxx_readHandler(comm);
      break;
    case BarrierFanInMsg:
      hpcxx_barrierFanIn(comm);
      break;
    case BarrierFanOutMsg:
      hpcxx_barrierFanOut(comm);
      break;
    case EagerPutMsg :
      hpcxx_writeHandler(comm);
      break;
    case EndOfFetchMsg:
      comm.signal();
      break;
    case RemoteFunctionMsg :
      hpcxx_invokeGFHandler(comm);
      break;
    case RemoteMethodMsg :
      hpcxx_invokeRMIHandler(comm);
      break;
    default:
      cerr << "HPCxx_CommCoord::handleMessage, unknown message type" << endl;
  }
}


//---- release ----------------------------------------------------------------
// Releases what had been an "in use" communications object.
//
// comm   -- the communications object to release
//-----------------------------------------------------------------------------
void hbp_CommCoord::release(HPCxx_Comm* comm)
{
  comm->inUse = false;
}
